import logging
from dataclasses import asdict, dataclass
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

import polars as pl
from polars.plugins import register_plugin_function
from polars.type_aliases import IntoExpr

from polars_ols.utils import build_expressions_from_patsy_formula, parse_into_expr

logger = logging.getLogger(__name__)

__all__ = [
    "compute_least_squares",
    "compute_recursive_least_squares",
    "compute_rolling_least_squares",
    "least_squares_from_formula",
    "predict",
    "OLSKwargs",
    "RLSKwargs",
    "RollingKwargs",
]


@dataclass
class OLSKwargs:
    """Specifies parameters relevant for regularized linear models: LASSO / Ridge / ElasticNet.

    Attributes:
        alpha: Regularization strength. Defaults to 0.0.
        l1_ratio: Mixing parameter for ElasticNet regularization (0 for Ridge, 1 for LASSO).
            Defaults to None (equivalent to Ridge regression).
        max_iter: Maximum number of iterations. Defaults to 1000 iterations.
        tol: Tolerance for convergence criterion. Defaults to 0.0001.
        positive: Whether to enforce non-negativity constraints on coefficients.
            Defaults to False (no constraint on coefficients).
    """

    alpha: Optional[float] = 0.0
    l1_ratio: Optional[float] = None
    max_iter: Optional[int] = 1_000
    tol: Optional[float] = 0.0001
    positive: Optional[bool] = False  # if True, imposes non-negativity constraint on coefficients

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


@dataclass
class RLSKwargs:
    """Specifies parameters of Recursive Least Squares models."""

    half_life: Optional[float] = None
    initial_state_covariance: Optional[float] = 10.0
    initial_state_mean: Optional[List[float] | float] = None

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


@dataclass
class RollingKwargs:
    """Specifies parameters of Rolling OLS model."""

    window_size: int = 1_000_000  # defaults to expanding OLS
    min_periods: Optional[int] = None
    use_woodbury: Optional[bool] = None
    alpha: Optional[float] = None  # optional ridge alpha

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


def _pre_process_data(
    target: pl.Expr, *features: pl.Expr, sample_weights: Optional[pl.Expr], add_intercept: bool
):
    """Pre-processes the input data by casting it to float32 and scaling it with sample weights if
     provided.

    Args:
        target: The target expression.
        *features: Variable number of feature expressions.
        sample_weights: Optional expression representing sample weights.
        add_intercept: Whether to add an intercept column.

    Returns:
        Tuple containing the pre-processed target, features, and sample weights.
    """
    target = parse_into_expr(target).cast(pl.Float32)
    features = [f.cast(pl.Float32) for f in features]
    if add_intercept:
        if any(f.meta.output_name == "const" for f in features):
            logger.info("feature named 'const' already detected, assuming it is an intercept")
        else:
            features += [target.mul(0.0).add(1.0).alias("const")]
    sqrt_w = 1.0
    if sample_weights is not None:
        sqrt_w = sample_weights.cast(pl.Float32).sqrt()
        target *= sqrt_w
        features = [expr * sqrt_w for expr in features]
    return target, features, sqrt_w


def compute_least_squares(
    target: IntoExpr,
    *features: pl.Expr,
    sample_weights: Optional[pl.Expr] = None,
    add_intercept: bool = False,
    mode: Literal["predictions", "residuals", "coefficients"] = "predictions",
    ols_kwargs: Optional[OLSKwargs] = None,
) -> pl.Expr:
    """Performs least squares regression.

    Depending on parameters provided this method supports a combination of sample weighting (WLS),
     L1/L2 regularization,
     and/or non-negativity constraint on coefficients.

    Args:
        target: The target expression.
        *features: Variable number of feature expressions.
        sample_weights: Optional expression representing sample weights.
        add_intercept: Whether to add an intercept column.
        mode: Mode of operation ("predictions", "residuals", "coefficients").
        ols_kwargs: Additional keyword arguments specific for regularized OLS models. These include:
            - "alpha": Regularization strength. Default is 0.0.
                      Expected dtype: float.
            - "l1_ratio": Mixing parameter for ElasticNet regularization (0 for Ridge, 1 for LASSO).
                          Default is None (equivalent to Ridge regression).
                          Expected dtype: float or None.
            - "max_iter": Maximum number of iterations. Defaults to 1000 iterations.
                          Expected dtype: int or None.
            - "tol": Tolerance for convergence criterion. Defaults to 0.0001.
                     Expected dtype: float or None.
            - "positive": Whether to enforce non-negativity constraints on coefficients.
                          Default is None (equivalent to False, i.e. no constraint on coefficients).
                          Expected dtype: bool or None.
    Returns:
        Resulting expression based on the chosen mode.
    """
    assert mode in {
        "predictions",
        "residuals",
        "coefficients",
    }, "'mode' must be one of {predictions, residuals, coefficients}"
    target, features, sqrt_w = _pre_process_data(
        target, *features, sample_weights=sample_weights, add_intercept=add_intercept
    )

    ols_kwargs: OLSKwargs = ols_kwargs or OLSKwargs()

    # register either coefficient or prediction plugin functions
    if mode == "coefficients":
        return (
            register_plugin_function(
                plugin_path=Path(__file__).parent,
                function_name="least_squares_coefficients",
                args=[target, *features],
                kwargs=ols_kwargs.to_dict(),
                is_elementwise=False,
                changes_length=True,
                returns_scalar=True,
            )
            .alias("coefficients")
            .struct.rename_fields([f.meta.output_name() for f in features])
        )
    else:
        predictions = (
            register_plugin_function(
                plugin_path=Path(__file__).parent,
                function_name="least_squares",
                args=[target, *features],
                kwargs=ols_kwargs.to_dict(),
                is_elementwise=False,
            )
            / sqrt_w
        )  # undo the sqrt(w) scaling implicit in predictions
        if mode == "predictions":
            return predictions
        else:
            return target - predictions


def compute_recursive_least_squares(
    target: IntoExpr,
    *features: pl.Expr,
    sample_weights: Optional[pl.Expr] = None,
    add_intercept: bool = False,
    mode: Literal["predictions", "residuals", "coefficients"] = "predictions",
    rls_kwargs: Optional[RLSKwargs] = None,
) -> pl.Expr:
    """Performs an efficient recursive least squares regression (RLS).

    Defaults to RLS with forgetting factor of 1.0 and a high initial state variance: equivalent to
     efficient 'streaming' expanding window OLS.

    Args:
        target: The target expression.
        *features: Variable number of feature expressions.
        sample_weights: Optional expression representing sample weights.
        add_intercept: Whether to add an intercept column.
        mode: Mode of operation ("predictions", "residuals", "coefficients").
        rls_kwargs: Additional keyword arguments for the recursive least squares model.
            - "half_life": Half-life parameter for exponential forgetting. Default is None
            (no forgetting).
            Expected dtype: float or None.
            - "initial_state_covariance":
            Scalar representing which behaves like an L2 regularization parameter. Larger values
             correspond to larger prior uncertainty around mean vector of state (inversely
             proportional to strength of equivalent L2 penalty).
            Defaults to 10.
            Expected dtype: float or None.
            - "initial_state_mean": Initial mean vector of the state.
                                    Default is zero vector.
                                    Expected dtype: List[float] or None.

    Returns:
        Resulting expression based on the chosen mode.
    """
    assert mode in {
        "predictions",
        "residuals",
        "coefficients",
    }, "'mode' must be one of {predictions, residuals, coefficients}"
    target, features, sqrt_w = _pre_process_data(
        target, *features, sample_weights=sample_weights, add_intercept=add_intercept
    )
    rls_kwargs: RLSKwargs = rls_kwargs or RLSKwargs()

    # register either coefficient or prediction plugin functions
    if mode == "coefficients":
        return (
            register_plugin_function(
                plugin_path=Path(__file__).parent,
                function_name="recursive_least_squares_coefficients",
                args=[target, *features],
                kwargs=rls_kwargs.to_dict(),
                is_elementwise=False,
            )
            .alias("coefficients")
            .struct.rename_fields([f.meta.output_name() for f in features])
        )
    else:
        predictions = (
            register_plugin_function(
                plugin_path=Path(__file__).parent,
                function_name="recursive_least_squares",
                args=[target, *features],
                kwargs=rls_kwargs.to_dict(),
                is_elementwise=False,
            )
            / sqrt_w
        )  # undo the sqrt(w) scaling implicit in predictions
        if mode == "predictions":
            return predictions
        else:
            return target - predictions


def compute_rolling_least_squares(
    target: IntoExpr,
    *features: pl.Expr,
    sample_weights: Optional[pl.Expr] = None,
    add_intercept: bool = False,
    mode: Literal["predictions", "residuals", "coefficients"] = "predictions",
    rolling_kwargs: Optional[RollingKwargs] = None,
) -> pl.Expr:
    """Performs least squares regression in a rolling window fashion.

    Args:
        target: The target expression.
        *features: Variable number of feature expressions.
        sample_weights: Optional expression representing sample weights.
        add_intercept: Whether to add an intercept column.
        mode: Mode of operation ("predictions", "residuals", "coefficients").
        rolling_kwargs: Additional keyword arguments for the rolling least squares model.
            - "window_size": The size of the rolling window.
            - "min_periods": The minimum number of observations required to produce estimates.
            - "use_woodbury": Whether to use Woodbury matrix identity for faster computation.
                              Defaults to True if num_features > 10.
            - "alpha": L2 Regularization strength. Default is 0.0.
                      Expected dtype: float.

    Returns:
        Resulting expression based on the chosen mode.
    """
    assert mode in {
        "predictions",
        "residuals",
        "coefficients",
    }, "'mode' must be one of {predictions, residuals, coefficients}"
    target, features, sqrt_w = _pre_process_data(
        target, *features, sample_weights=sample_weights, add_intercept=add_intercept
    )
    rolling_kwargs: RollingKwargs = rolling_kwargs or RollingKwargs()

    # register either coefficient or prediction plugin functions
    if mode == "coefficients":
        return (
            register_plugin_function(
                plugin_path=Path(__file__).parent,
                function_name="rolling_least_squares_coefficients",
                args=[target, *features],
                kwargs=rolling_kwargs.to_dict(),
                is_elementwise=False,
            )
            .alias("coefficients")
            .struct.rename_fields([f.meta.output_name() for f in features])
        )
    else:
        predictions = (
            register_plugin_function(
                plugin_path=Path(__file__).parent,
                function_name="rolling_least_squares",
                args=[target, *features],
                kwargs=rolling_kwargs.to_dict(),
                is_elementwise=False,
            )
            / sqrt_w
        )  # undo the sqrt(w) scaling implicit in predictions
        if mode == "predictions":
            return predictions
        else:
            return target - predictions


def least_squares_from_formula(
    formula: str,
    sample_weights: Optional[pl.Expr] = None,
    mode: Literal["predictions", "residuals", "coefficients"] = "predictions",
    **kwargs,
) -> pl.Expr:
    """Performs least squares regression using a formula.

    Depending on choice of additional kwargs dispatches either rolling, recursive,
    on static least squares compute functions.

    Args:
        formula: Patsy-style formula string.
        **kwargs: Additional keyword arguments for the least squares function.

    Returns:
        Resulting expression based on the formula.
    """
    expressions, add_intercept = build_expressions_from_patsy_formula(
        formula, include_dependent_variable=True
    )

    # resolve additional kwargs and relevant ols compute function
    if kwargs.get("half_life"):
        rls_kwargs: RLSKwargs = RLSKwargs(**kwargs)
        func = partial(compute_recursive_least_squares, rls_kwargs=rls_kwargs)
    elif kwargs.get("window_size"):
        rolling_kwargs: RollingKwargs = RollingKwargs(**kwargs)
        func = partial(compute_rolling_least_squares, rolling_kwargs=rolling_kwargs)
    else:
        ols_kwargs: OLSKwargs = OLSKwargs(**kwargs)
        func = partial(compute_least_squares, ols_kwargs=ols_kwargs)

    return func(
        expressions[0],
        *expressions[1:],
        add_intercept=add_intercept,
        sample_weights=sample_weights,
        mode=mode,
    )


def predict(
    coefficients: IntoExpr,
    *features: pl.Expr,
    name: Optional[str] = None,
    add_intercept: bool = False,
) -> pl.Expr:
    """Helper which computes predictions as a product of (aligned) coefficients with features.

    Args:
        coefficients: Polars expression returning a coefficients struct.
        *features: variable number of feature expressions.
        add_intercept: boolean indicating if a constant should be added to features.

    Returns:
        polars expression denoting computed predictions.
    """
    if add_intercept:
        if any(f.meta.output_name == "const" for f in features):
            logger.warning("feature named 'const' already detected, assuming it is the intercept")
        else:
            features += (features[-1].mul(0.0).add(1.0).alias("const"),)

    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="predict",
        args=[coefficients, *(f.cast(pl.Float32) for f in features)],
        is_elementwise=False,
    ).alias(name or "predictions")



from __future__ import annotations

from functools import lru_cache, reduce
from typing import TYPE_CHECKING, Sequence, Tuple

import polars as pl

if TYPE_CHECKING:
    from polars.type_aliases import IntoExpr, PolarsDataType


def parse_into_expr(
    expr: IntoExpr,
    *,
    str_as_lit: bool = False,
    list_as_lit: bool = True,
    dtype: PolarsDataType | None = None,
) -> pl.Expr:
    """
    Parse a single input into an expression.

    Parameters
    ----------
    expr
        The input to be parsed as an expression.
    str_as_lit
        Interpret string input as a string literal. If set to `False` (default),
        strings are parsed as column names.
    list_as_lit
        Interpret list input as a lit literal, If set to `False`,
        lists are parsed as `Series` literals.
    dtype
        If the input is expected to resolve to a literal with a known dtype, pass
        this to the `lit` constructor.

    Returns
    -------
    polars.Expr
    """
    if isinstance(expr, pl.Expr):
        pass
    elif isinstance(expr, str) and not str_as_lit:
        expr = pl.col(expr)
    elif isinstance(expr, list) and not list_as_lit:
        expr = pl.lit(pl.Series(expr), dtype=dtype)
    else:
        expr = pl.lit(expr, dtype=dtype)

    return expr


@lru_cache(maxsize=100)
def build_expressions_from_patsy_formula(
    formula: str, include_dependent_variable: bool = False
) -> Tuple[Sequence[pl.Expr], bool]:
    """Builds polars LHS and/or RHS expressions given a patsy formula.

    Only a subset of supported features are supported, in particular:
    - simple target to feature formula w/ interaction variables and intercept are
    fully supported (e.g. 'y ~ x1 + x2:x3 -1')
    - external functions applied to columns are not yet supported (e.g. "log(x1)")
    - categorical features can not be supported, you have to 'pivot' yourself (e.g. "C(group)")

    Example:
        >>> ex, intercept = build_expressions_from_patsy_formula("y ~ x1 + x2 + x3:x4",
        ... include_dependent_variable=True)
        >>> assert str(pl.col("y")) == str(ex[0])
        >>> assert str(pl.col("x1")) == str(ex[1])
        >>> assert str((1 * pl.col("x3") * pl.col("x4")).alias("x3:x4")) == str(ex[-1])
    """
    try:
        import patsy as pa
    except ImportError as e:
        raise NotImplementedError(
            "'patsy' needs to be installed in your python environment in order to use "
            "formula api"
        ) from e
    desc = pa.ModelDesc.from_formula(formula)

    if include_dependent_variable:
        assert len(desc.lhs_termlist) == 1, "must provide exactly one LHS variable"
        terms = desc.lhs_termlist + desc.rhs_termlist
    else:
        assert len(desc.lhs_termlist) == 0, "can not provide LHS variables in this context"
        terms = desc.rhs_termlist

    add_intercept: bool = "-1" not in formula

    expressions = []
    for term in terms:
        if any("C(" in f.code for f in term.factors):
            raise NotImplementedError(
                "building patsy categories into polars expressions is not supported"
            )
        if len(term.factors) == 1:
            expressions.append(pl.col(term.factors[0].code))
        elif len(term.factors) >= 2:
            expr = reduce((lambda x, y: x * pl.col(y)), (f.code for f in term.factors), pl.lit(1))
            expressions.append(expr.alias(":".join(f.code for f in term.factors)))
    return expressions, add_intercept



import time
from contextlib import contextmanager

import numpy as np
import polars as pl
import statsmodels.formula.api as smf
from sklearn.linear_model import ElasticNet
from statsmodels.regression.rolling import RollingOLS

from polars_ols import OLSKwargs, compute_least_squares, least_squares_from_formula


@contextmanager
def timer(msg: str | None = None, precision: int = 3) -> float:
    start = time.perf_counter()
    end = start
    yield lambda: end - start
    msg = f"{msg or 'Took'}: {time.perf_counter() - start:.{precision}f} seconds"
    print(msg)
    end = time.perf_counter()


def _make_data(n: int = 10_000, n_groups: int | None = None):
    rng = np.random.default_rng(0)
    array = rng.normal(size=(n, 2))
    data = {
        "y": array.sum(axis=1) + rng.normal(size=n, scale=0.1),
        "x1": array[:, 0],
        "x2": array[:, 1],
    }
    if n_groups is not None:
        data |= {"group": rng.integers(n_groups, size=n)}
    return pl.DataFrame(data).with_columns(pl.col(pl.FLOAT_DTYPES).cast(pl.Float32))


def test_ols():
    df = _make_data()
    # compute OLS solution w/ polars (via QR in rust) [2.331s]
    with timer("OLS rust"):
        for _ in range(1_000):
            expr = compute_least_squares(pl.col("y"), pl.col("x1"), pl.col("x2")).alias(
                "predictions"
            )
            df = df.lazy().with_columns(expr).collect()

    # compute OLS w/ lstsq numpy [4.583s]
    with timer("OLS numpy"):
        for _ in range(1_000):
            x, y = df.select("x1", "x2").to_numpy(), df.select("y").to_numpy().flatten()
            coef = np.linalg.lstsq(x, y, rcond=None)[0]
            df = df.with_columns(predictions2=pl.lit(x @ coef).flatten())

    assert np.allclose(df["predictions"], df["predictions2"], atol=1.0e-4, rtol=1.0e-4)


def test_coefficients_ols():
    df = _make_data()
    coef = (
        df.select(
            pl.col("y")
            .least_squares.from_formula("x1 + x2 -1", mode="coefficients")
            .alias("coefficients")
        )
        .unnest("coefficients")
        .to_numpy()[0]
    )
    assert np.allclose(coef, [1.0, 1.0], atol=1.0e-2, rtol=1.0e-2)


def test_coefficients_ols_groups():
    df = _make_data(n_groups=10)
    coef_group = (
        df.select(
            "group",
            pl.col("y")
            .least_squares.from_formula("x1 + x2 -1", mode="coefficients")
            .over("group")
            .alias("coefficients"),
        )
        .unique()
        .unnest("coefficients")
    )
    assert len(coef_group) == 10

    coef_group_1 = (
        df.filter(pl.col("group") == 1)
        .select(
            pl.col("y")
            .least_squares.from_formula("x1 + x2 -1", mode="coefficients")
            .alias("coefficients")
        )
        .unnest("coefficients")
    )
    assert np.allclose(coef_group.filter(pl.col("group") == 1).select("x1", "x2"), coef_group_1)


def test_coefficients_shape_broadcast():
    df = _make_data(n=10_000, n_groups=10)
    assert df.select(
        pl.col("y")
        .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
        .alias("coefficients")
    ).shape == (1, 1)

    assert df.with_columns(
        pl.col("y")
        .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
        .alias("coefficients")
    ).shape == (10_000, 5)

    df_group = df.select(
        pl.col("y")
        .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
        .over("group")
        .alias("coefficients"),
        "group",
    )
    assert df_group.shape == (10_000, 2)
    assert df_group.unique().shape == (10, 2)

    assert df.with_columns(
        pl.col("y")
        .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
        .over("group")
        .alias("coefficients")
    ).shape == (10_000, 5)


def test_ols_residuals():
    df = _make_data()
    residuals = df.select(
        pl.col("y").least_squares.from_formula("x1 + x2 -1", mode="residuals")
    ).to_numpy()
    x, y = df.select("x1", "x2").to_numpy(), df["y"].to_numpy()
    coef = np.linalg.lstsq(x, y, rcond=None)[0]
    assert np.allclose(residuals.flatten(), y - x @ coef, rtol=1.0e-4, atol=1.0e-4)


def test_ols_intercept():
    df = _make_data()
    expr = compute_least_squares(pl.col("y"), pl.col("x1"), pl.col("x2"), add_intercept=True).alias(
        "predictions"
    )
    y_hat = df.select(expr).to_numpy().flatten()
    expected = smf.ols(formula="y ~ x1 + x2", data=df).fit().predict(df).to_numpy()
    assert np.allclose(y_hat, expected, atol=1.0e-4, rtol=1.0e-4)


def test_least_squares_from_formula():
    weights = np.random.uniform(0, 1, size=10_000)
    weights /= weights.mean()
    df = _make_data().with_columns(sample_weights=pl.lit(weights)).cast(pl.Float32)

    expr = least_squares_from_formula(
        "y ~ x1 + x2",  # patsy includes intercept by default
        sample_weights=pl.col("sample_weights"),
    ).alias("predictions")

    expected = (
        smf.wls("y ~ x1 + x2", data=df, weights=df["sample_weights"].to_numpy())
        .fit()
        .predict(df)
        .to_numpy()
    )
    assert np.allclose(df.select(expr).to_numpy().flatten(), expected, rtol=1.0e-4, atol=1.0e-4)


def test_ridge():
    df = _make_data()
    alpha: float = 0.01

    with timer("ridge python"):
        for _ in range(1_000):
            x = df.select("x1", "x2").to_numpy()
            y = df.select("y").to_numpy().flatten()
            coef_expected = np.linalg.solve((x.T @ x) + np.eye(x.shape[1]) * alpha, x.T @ y)
            expected = x @ coef_expected

    with timer("ridge rust"):
        for _ in range(1_000):
            expr = compute_least_squares(
                pl.col("y"),
                pl.col("x1"),
                pl.col("x2"),
                ols_kwargs=OLSKwargs(alpha=alpha),
            ).alias("predictions")
            df = df.lazy().with_columns(expr).collect()
    assert np.allclose(df["predictions"].to_numpy(), expected, rtol=1.0e-4, atol=1.0e-4)


def test_wls():
    array = np.random.normal(size=(10_000, 2))
    df = pl.DataFrame(
        {
            "y": array.sum(axis=1)
            + np.hstack(
                [np.random.normal(size=8_000, scale=10.0), np.random.normal(size=2_000, scale=0.1)]
            ),
            "x1": array[:, 0],
            "x2": array[:, 1],
        }
    ).cast(pl.Float32)

    weights = np.hstack([np.ones(8_000) * 1.0 / 10**2, np.ones(2_000) * 1.0 / 0.1**2])
    weights /= weights.mean()

    df = df.with_columns(sample_weight=weights).cast(pl.Float32)

    expr_wls = compute_least_squares(
        pl.col("y"),
        pl.col("x1"),
        pl.col("x2"),
        sample_weights=pl.col("sample_weight"),
    ).alias("predictions_wls")
    expr_ols = compute_least_squares(
        pl.col("y"),
        pl.col("x1"),
        pl.col("x2"),
    ).alias("predictions_ols")
    df = df.lazy().with_columns(expr_wls, expr_ols).collect()

    y_hat_wls = smf.wls(data=df, formula="y ~ x1 + x2 -1", weights=weights).fit().predict()
    y_hat_ols = smf.ols(data=df, formula="y ~ x1 + x2 -1").fit().predict()

    assert np.allclose(df["predictions_ols"].to_numpy(), y_hat_ols, rtol=1.0e-4, atol=1.0e-4)
    assert np.allclose(df["predictions_wls"].to_numpy(), y_hat_wls, rtol=1.0e-4, atol=1.0e-4)


def test_least_squares_namespace():
    df = _make_data().with_columns(sample_weight=pl.lit(1.0))
    alpha: float = 0.0
    ols = pl.col("y").least_squares.ols(pl.col("x1"), pl.col("x2")).alias("ols")
    ridge = pl.col("y").least_squares.ridge(pl.col("x1"), pl.col("x2"), alpha=alpha).alias("ridge")
    wls = (
        pl.col("y")
        .least_squares.wls(pl.col("x1"), pl.col("x2"), sample_weights=pl.col("sample_weight"))
        .alias("wls")
    )
    formula = pl.col("y").least_squares.from_formula("x1 + x2 - 1").alias("formula")
    df = df.lazy().select(ols, ridge, wls, formula).collect()

    # ensure all of the above are equivalent
    assert np.allclose(df.corr(), 1.0)


def test_elastic_net():
    df = _make_data()
    mdl = ElasticNet(fit_intercept=False, alpha=0.1, l1_ratio=0.5, max_iter=1_000, tol=0.0001)
    mdl.fit(df.select(pl.all().exclude("y")), df.select("y"))

    coef = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.from_formula(
                "x1 + x2 -1",
                mode="coefficients",
                l1_ratio=0.5,
                alpha=0.1,
                max_iter=1_000,
                tol=0.0001,
            )
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
        .flatten()
    )
    assert np.allclose(mdl.coef_, coef, rtol=1.0e-4, atol=1.0e-4)


def test_elastic_net_non_negative():
    df = _make_data()
    mdl = ElasticNet(
        fit_intercept=False, alpha=0.1, l1_ratio=0.5, max_iter=1_000, tol=0.0001, positive=True
    )
    mdl.fit(df.select(pl.col("x1"), -pl.col("x2")), df.select("y"))

    coef = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.elastic_net(
                pl.col("x1"),
                -pl.col("x2"),
                mode="coefficients",
                l1_ratio=0.5,
                alpha=0.1,
                max_iter=1_000,
                tol=0.0001,
                positive=True,
            )
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
        .flatten()
    )
    assert np.allclose(mdl.coef_, coef, rtol=1.0e-4, atol=1.0e-4)


def test_recursive_least_squares():
    df = _make_data()

    # expanding OLS
    coef_rls = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.rls(
                pl.col("x1"),
                pl.col("x2"),
                mode="coefficients",
                # equivalent to expanding window (no forgetting)
                half_life=None,
                # arbitrarily weak L2 (diffuse) prior
                initial_state_covariance=1_000_000.0,
            )
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
    )

    # full sample OLS
    coef_ols = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
        .flatten()
    )

    # the two models should be equivalent in full sample
    assert np.allclose(coef_rls[-1], coef_ols, rtol=1.0e-4, atol=1.0e-4)


def test_recursive_least_squares_prior():
    df = _make_data()

    # equivalent to expanding OLS w/ L2 prior towards [0.5, 0.5]
    coef_rls_prior = (
        df.lazy()
        .select(
            pl.col("y")
            .least_squares.rls(
                pl.col("x1"),
                pl.col("x2"),
                mode="coefficients",
                # equivalent to expanding window (no forgetting)
                half_life=None,
                initial_state_covariance=1.0e-6,  # arbitrarily strong L2 prior
                initial_state_mean=[0.25, 0.25],  # custom prior
            )
            .alias("coefficients")
        )
        .unnest("coefficients")
        .collect()
        .to_numpy()
    )

    # given few samples and strong prior strength, the coefficients are nearly
    # identical to the prior
    assert np.allclose(coef_rls_prior[0], [0.25, 0.25], rtol=1.0e-3, atol=1.0e-3)
    assert np.allclose(coef_rls_prior[10], [0.25, 0.25], rtol=1.0e-3, atol=1.0e-3)

    # as number of samples seen grows, the coefficients start to diverge from prior
    # & eventually converge to ground truth.
    assert not np.allclose(coef_rls_prior[-1], [0.5, 0.5], rtol=1.0e-4, atol=1.0e-4)


def test_rolling_least_squares():
    df = _make_data()
    with timer("rolling ols"):
        coef_rolling = (
            df.lazy()
            .select(
                pl.col("y")
                .least_squares.rolling_ols(
                    pl.col("x1"),
                    pl.col("x2"),
                    mode="coefficients",
                    window_size=252,
                    min_periods=2,
                )
                .alias("coefficients")
            )
            .unnest("coefficients")
            .collect()
            .to_numpy()
        )
    with timer("rolling ols statsmodels"):
        mdl = RollingOLS(
            df["y"].to_numpy(), df[["x1", "x2"]].to_numpy(), window=252, min_nobs=2, expanding=True
        ).fit()
    assert np.allclose(coef_rolling[1:], mdl.params[1:].astype("float32"), rtol=1.0e-3, atol=1.0e-3)


def test_moving_window_regressions_over():
    df = _make_data(n_groups=10)

    df = (
        (
            df.lazy().select(
                "group",
                pl.col("y")
                .least_squares.rolling_ols(
                    pl.col("x1"),
                    pl.col("x2"),
                    mode="coefficients",
                    window_size=1_000_000,
                    min_periods=2,
                    # larger than data window size equivalent to expanding window
                )
                .over("group")
                .alias("coef_rolling_ols_group"),
                pl.col("y")
                .least_squares.rls(
                    pl.col("x1"),
                    pl.col("x2"),
                    half_life=None,
                    initial_state_covariance=1.0e6,
                    mode="coefficients",
                    # no forgetting factor + diffuse prior equivalent to expanding window
                )
                .over("group")
                .alias("coef_rls_group"),
                pl.col("y")
                .least_squares.ols(
                    pl.col("x1"),
                    pl.col("x2"),
                    mode="coefficients",
                )
                .over("group")
                .alias("coef_ols_group"),  # full sample OLS per group
            )
        )
        .collect()
        .rechunk()
    )

    # As of the last sample per group: RLS & rolling regression should behave identically
    # to full sample OLS per group (the way they were set up above)
    df_last = df.group_by("group").last()

    assert np.allclose(
        df_last.unnest("coef_ols_group").select("x1", "x2"),
        df_last.unnest("coef_rolling_ols_group").select("x1", "x2"),
    )
    assert np.allclose(
        df_last.unnest("coef_ols_group").select("x1", "x2"),
        df_last.unnest("coef_rls_group").select("x1", "x2"),
        rtol=1.0e-4,
        atol=1.0e-4,
    )


def test_predict():
    df = _make_data(n_groups=1)
    df_test = _make_data(n=20, n_groups=1)

    # estimate coefficients
    df_coefficients = (
        df.lazy()
        .select(
            "group",
            pl.col("y")
            .least_squares.ols(pl.col("x1"), pl.col("x2"), mode="coefficients")
            .over("group")
            .alias("coefficients"),
        )
        .unique()
    )

    # compute predictions as a rust expression
    predictions = (
        df_test.lazy()
        .join(df_coefficients, on="group")
        .select(
            pl.col("coefficients").least_squares.predict(
                pl.col("x1"), pl.col("x2"), name="predictions"
            )
        )
        .collect()
        .to_numpy()
    )

    # compare to predictions computed as dot product of train coefficients with test data
    expected = (
        df_test.drop("group")
        .to_numpy()
        .dot(df_coefficients.unnest("coefficients").collect().to_numpy().T)
    )
    assert np.allclose(predictions, expected)


def test_predict_formula():
    df = _make_data()
    df = (
        df.lazy()
        .with_columns(
            coefficients=pl.col("y").least_squares.from_formula("x1 + x2", mode="coefficients"),
            predictions_1=pl.col("y").least_squares.from_formula("x1 + x2", mode="predictions"),
        )
        .with_columns(
            predictions_2=pl.col("coefficients").least_squares.predict_from_formula("x1 + x2")
        )
    ).collect()
    assert np.allclose(df["predictions_1"], df["predictions_2"])


def test_predict_complex():
    df = _make_data(n_groups=10)
    df = (
        df.lazy()
        .with_columns(
            predictions_1=pl.col("y")
            .least_squares.rls(pl.col("x1"), pl.col("x2"), mode="predictions")
            .over("group"),
            coefficients=pl.col("y")
            .least_squares.rls(pl.col("x1"), pl.col("x2"), mode="coefficients")
            .over("group"),
        )
        .with_columns(
            predictions_2=pl.col("coefficients").least_squares.predict(pl.col("x1"), pl.col("x2"))
        )
        .collect()
    )
    assert np.allclose(df["predictions_1"], df["predictions_2"])



import numpy as np
import polars as pl
import pyperf
import statsmodels.formula.api as smf
from sklearn.linear_model import ElasticNet
from statsmodels.regression.recursive_ls import RecursiveLS
from statsmodels.regression.rolling import RollingOLS

import polars_ols as pls  # import package to register the .least_squares namespace


def _make_data(n_samples: int = 2_000, n_features: int = 5) -> pl.DataFrame:
    x = np.random.normal(size=(n_samples, n_features)).astype("float32")
    eps = np.random.normal(size=n_samples, scale=0.1).astype("float32")
    return pl.DataFrame(data=x, schema=[f"x{i + 1}" for i in range(n_features)]).with_columns(
        y=pl.lit(x.sum(1) + eps)
    )


def benchmark_least_squares(data: pl.DataFrame):
    return (
        data.lazy()
        .with_columns(
            pl.col("y")
            .least_squares.ols(*[pl.col(c) for c in data.columns if c != "y"])
            .alias("predictions")
        )
        .collect()
    )


def benchmark_least_squares_numpy(data: pl.DataFrame):
    y, x = data.select("y").to_numpy().flatten(), data.select(pl.all().exclude("y")).to_numpy()
    coef = np.linalg.lstsq(x, y, rcond=None)[0]
    return data.lazy().with_columns(predictions=pl.lit(x @ coef).alias("predictions")).collect()


def benchmark_ridge(data: pl.DataFrame):
    return (
        data.lazy()
        .with_columns(
            pl.col("y")
            .least_squares.ridge(*[pl.col(c) for c in data.columns if c != "y"], alpha=0.0001)
            .alias("predictions")
        )
        .collect()
    )


def benchmark_ridge_numpy(data: pl.DataFrame):
    alpha: float = 0.0001
    y, x = data.select("y").to_numpy().flatten(), data.select(pl.all().exclude("y")).to_numpy()
    xtx = x.T @ x + np.eye(x.shape[1]) * alpha
    xty = x.T @ y
    coef = np.linalg.solve(xtx, xty)
    return data.lazy().with_columns(predictions=pl.lit(x @ coef).alias("predictions")).collect()


def benchmark_wls_from_formula(data: pl.DataFrame):
    return (
        data.lazy()
        .with_columns(pl.lit(1.0).alias("sample_weights"))
        .with_columns(
            pl.col("y")
            .least_squares.from_formula(
                " + ".join(c for c in data.columns if "x" in c),
                sample_weights=pl.col("sample_weights"),
            )
            .alias("predictions")
        )
    ).collect()


def benchmark_wls_from_formula_statsmodels(data: pl.DataFrame):
    x_cols = " + ".join(c for c in data.columns if "x" in c)
    predictions = (
        smf.wls(
            data=data,
            formula=f"y ~ {x_cols}",
            weights=np.ones(len(data), dtype="float32"),
        )
        .fit()
        .predict()
    )
    return data.lazy().with_columns(predictions=pl.lit(predictions)).collect()


def benchmark_elastic_net(data: pl.DataFrame):
    return (
        data.lazy()
        .with_columns(
            pl.col("y").least_squares.elastic_net(
                *[pl.col(c) for c in data.columns if c != "y"],
                alpha=0.1,
                l1_ratio=0.5,
                max_iter=1_000,
            )
        )
        .collect()
    )


def benchmark_elastic_net_sklearn(data: pl.DataFrame):
    mdl = ElasticNet(fit_intercept=False, alpha=0.1, l1_ratio=0.5, max_iter=1_000)
    y, x = data.select("y").to_numpy().flatten(), data.select(pl.all().exclude("y")).to_numpy()
    mdl.fit(x, y)
    return data.lazy().with_columns(predictions=pl.lit(mdl.predict(x))).collect()


def benchmark_recursive_least_squares(data: pl.DataFrame):
    return (
        data.lazy()
        .with_columns(
            pl.col("y").least_squares.rls(
                *[pl.col(c) for c in data.columns if c != "y"],
                half_life=252,
            )
        )
        .collect()
    )


def benchmark_rolling_least_squares(data: pl.DataFrame):
    features = [pl.col(c) for c in data.columns if c != "y"]
    return (
        data.lazy()
        .with_columns(
            pl.col("y").least_squares.rolling_ols(
                *features,
                window_size=252,
                min_periods=len(features),
            )
        )
        .collect()
    )


def benchmark_rolling_least_squares_statsmodels(data: pl.DataFrame):
    x = data.select(pl.all().exclude("y")).to_numpy()
    res = RollingOLS(df["y"].to_numpy(), x, window=252, min_nobs=x.shape[1], expanding=True).fit(
        params_only=True
    )
    return data.lazy().with_columns(predictions=pl.lit((res.params * x).sum(1))).collect()


def benchmark_recursive_least_squares_statsmodels(data: pl.DataFrame):
    x = data.select(pl.all().exclude("y")).to_numpy()
    res = RecursiveLS(
        df["y"].to_numpy(),
        x,
    ).fit(params_only=True)
    return data.lazy().with_columns(predictions=pl.lit((res.params * x).sum(1))).collect()


if __name__ == "__main__":
    # example: python tests/benchmark.py --quiet --fast
    # we run the benchmarks in python (as opposed to rust) so that overhead of pyO3 is included
    df = _make_data(n_features=10, n_samples=2_000)
    runner = pyperf.Runner()
    runner.bench_func("benchmark_least_squares", benchmark_least_squares, df)
    runner.bench_func("benchmark_ridge", benchmark_ridge, df)
    runner.bench_func("benchmark_wls_from_formula", benchmark_wls_from_formula, df)
    runner.bench_func("benchmark_elastic_net", benchmark_elastic_net, df)
    runner.bench_func("benchmark_recursive_least_squares", benchmark_recursive_least_squares, df)
    runner.bench_func("benchmark_rolling_least_squares", benchmark_rolling_least_squares, df)

    # runner.bench_func("benchmark_least_squares_numpy", benchmark_least_squares_numpy, df)
    # runner.bench_func("benchmark_ridge_numpy", benchmark_ridge_numpy, df)
    # runner.bench_func(
    #     "benchmark_wls_from_formula_statsmodels", benchmark_wls_from_formula_statsmodels, df
    # )
    # runner.bench_func("benchmark_elastic_net_sklearn", benchmark_elastic_net_sklearn, df)
    # runner.bench_func(
    #     "benchmark_recursive_least_squares_statsmodels",
    #     benchmark_rolling_least_squares_statsmodels,
    #     df,
    # )
    # runner.bench_func(
    #     "benchmark_rolling_least_squares_statsmodels",
    #     benchmark_rolling_least_squares_statsmodels,
    #     df,
    # )



from typing import Literal, Optional

import polars as pl

from polars_ols.least_squares import (
    OLSKwargs,
    RLSKwargs,
    RollingKwargs,
    compute_least_squares,
    compute_recursive_least_squares,
    compute_rolling_least_squares,
    least_squares_from_formula,
    predict,
)
from polars_ols.utils import build_expressions_from_patsy_formula

__all__ = [
    "compute_least_squares",
    "compute_recursive_least_squares",
    "compute_rolling_least_squares",
    "LeastSquares",
    "least_squares_from_formula",
]


@pl.api.register_expr_namespace("least_squares")
class LeastSquares:
    def __init__(self, expr: pl.Expr):
        self._expr = expr

    def least_squares(
        self,
        *features: pl.Expr,
        sample_weights: Optional[pl.Expr] = None,
        add_intercept: bool = False,
        mode: Literal["predictions", "residuals", "coefficients"] = "predictions",
        **ols_kwargs,
    ) -> pl.Expr:
        return compute_least_squares(
            self._expr,
            *features,
            sample_weights=sample_weights,
            add_intercept=add_intercept,
            mode=mode,
            ols_kwargs=OLSKwargs(**ols_kwargs),
        )

    def ols(self, *features: pl.Expr, **kwargs) -> pl.Expr:
        return self.least_squares(*features, **kwargs)

    def wls(self, *features: pl.Expr, sample_weights: pl.Expr, **kwargs) -> pl.Expr:
        return self.least_squares(*features, sample_weights=sample_weights, **kwargs)

    def ridge(self, *features: pl.Expr, alpha: float, **kwargs) -> pl.Expr:
        return self.least_squares(*features, alpha=alpha, l1_ratio=0.0, **kwargs)

    def lasso(self, *features: pl.Expr, alpha: float, **kwargs) -> pl.Expr:
        return self.least_squares(*features, alpha=alpha, l1_ratio=1.0, **kwargs)

    def elastic_net(self, *features: pl.Expr, alpha: float, l1_ratio: float = 0.5, **kwargs):
        return self.least_squares(*features, alpha=alpha, l1_ratio=l1_ratio, **kwargs)

    def rls(
        self,
        *features: pl.Expr,
        sample_weights: Optional[pl.Expr] = None,
        add_intercept: bool = False,
        mode: Literal["predictions", "residuals", "coefficients"] = "predictions",
        **rls_kwargs,
    ):
        return compute_recursive_least_squares(
            self._expr,
            *features,
            sample_weights=sample_weights,
            add_intercept=add_intercept,
            mode=mode,
            rls_kwargs=RLSKwargs(**rls_kwargs),
        )

    def rolling_ols(
        self,
        *features: pl.Expr,
        sample_weights: Optional[pl.Expr] = None,
        add_intercept: bool = False,
        mode: Literal["predictions", "residuals", "coefficients"] = "predictions",
        **rolling_kwargs,
    ):
        return compute_rolling_least_squares(
            self._expr,
            *features,
            sample_weights=sample_weights,
            add_intercept=add_intercept,
            mode=mode,
            rolling_kwargs=RollingKwargs(**rolling_kwargs),
        )

    def expanding_ols(self, *features: pl.Expr, **kwargs):
        return self.rls(*features, half_life=None, **kwargs)

    def from_formula(self, formula: str, **kwargs) -> pl.Expr:
        features, add_intercept = build_expressions_from_patsy_formula(
            formula, include_dependent_variable=False
        )
        if kwargs.get("half_life"):
            return self.rls(*features, add_intercept=add_intercept, **kwargs)
        elif kwargs.get("window_size"):
            return self.rolling_ols(*features, add_intercept=add_intercept, **kwargs)
        else:
            return self.least_squares(*features, add_intercept=add_intercept, **kwargs)

    def predict(
        self, *features: pl.Expr, name: Optional[str] = None, add_intercept: bool = False
    ) -> pl.Expr:
        return predict(self._expr, *features, name=name, add_intercept=add_intercept)

    def predict_from_formula(self, formula: str, name: Optional[str] = None) -> pl.Expr:
        features, add_intercept = build_expressions_from_patsy_formula(
            formula, include_dependent_variable=False
        )
        has_const = any(f.meta.output_name == "const" for f in features)
        add_intercept &= not has_const
        return self.predict(*features, add_intercept=add_intercept, name=name)
